#include <stdio.h>   /* for printf() */
#include <stdlib.h>
#include <string.h>

#include "music.h"
#include "m_irq.h"
#include "m_data.h" /* for Period[] */
#include "m_sr.h"
#include "bitfun.h"
#include <math.h>

#ifdef DJGPP
#define printf(...)
#endif

#ifdef linux
#define PITCHSENSE 1000 /* default value: 200 */
#define SEMIUNIT   100
#else
#define PITCHSENSE 127 /* default value: 2 */
#define SEMIUNIT   1
#endif

/* 
 Supported commandinfobytes currently include:
 
  Axx (set speed (ticks))
  Bxx
  Cxx
  Dx0 (volume slide up   ticks*x)
  D0y (volume slide down ticks*y)
  DxF (increase volume by y)
  DFy (decrease volume by y)
  DFF (maximizes volume)

  EFx (freqdec)    (finally correct!! 2000-09-21 / 3.0.2)
  EEx (UNTESTED)   (extra fine slide down)
  Exx (freqslide)  (-"-)

  FFx (freqinc)    (-"-)
  FEx (UNTESTED)   (extra fine slide up)
  Fxx (freqslide)  (-"-)

  Hxx (nearly correct 2000-09-21 - at least it works on adlib)
  Kxx (-"-)
  Uxx (fine vibrato, UNTESTED. Where is this documented?)
  
  Ixy (Tremor, UNTESTED. Ontime x, offtime y)
  
  Rxy (Tremolo, UNTESTED)

  Gxx (Should work on adlib. *UNTESTED*. / 3.0.2)
  Lxx (-"-)

  Jxx (arpeggio)
  Pxx (NON-STANDARD! Specify arpeggio speed - default=32h (50Hz))
      (If bit 7 set, then arpeggio is only with two notes, not three.)

  Xxx (panning, quite grain in FM... 00..FF, 80h=middle)

  Qxy (BETA)

  SBx (patternloop, x>0?count:start)
  SCx (notecut in   x ticks)
  SDx (notedelay    x ticks)
  SEx (patterndelay x lines)

  MBx (NON-STANDARD: reverb)
  MCx (NON-STANDARD: chorus)

  Txx (set tempo)

  By definition, rows played in minute is 24 * tempo / ticks.
  That means, in a second, it is 2/5*tempo/ticks.
  Example: T7D, A06: tick length is 5 / (125+125) = 1 / 50 seconds,
                     six ticks in row, means row is 6 / 50 seconds,
                     25/3 = 8.333.. rows per second.
  Timer ticks per row = rate * speed * 5 / tempo / 2

  Exx and Fxx:

    For Exx, PeriodSlide =  xx
    For Fxx, PeriodSlide = -xx

    In each frame (there are Axx frames in a row)
      the Period of note will be increased by PeriodSlide.

  Hxy and Kxy:
  
    For Kxy: first do Dxy and then set xy to be the last Hxy xy.

    Vibrato amplitude: y/16       semitones
    Vibrato speed:     x*ticks/64 cycles in second
    Ticks is the value set by Axx -command.
    
    Uxy is finer than Hxy.
  
  Gxx and Lxy:
  
    For Lxy: first do Dxy and then set xx to be the last Gxx xx.
    
    Tune the period by xx in each tick, until the period is correct.

  To fix:

    SEx (does not appear to handle command repeatings correctly)
    SBx (this has broken in some way)

*/

volatile struct MusData MusData;

static unsigned MakePeriod(unsigned Note, unsigned Oct, unsigned C4Spd)
{
    return (unsigned long)8363 * 16 * Period[11-Note] / C4Spd >> Oct;
}

static void M_NoteOff(int ch)
{
    MusData.driver->noteoff(ch);
}

static void M_MidiCtrl(int ch, int ctrl, int value)
{
    MusData.driver->midictrl(ch, ctrl, value);
}

static void M_Pan(int ch, int value)
{
    /* Pan limits: -128..+127 */
    MusData.driver->setpan(ch, value - 128);
}

static void M_GmBend(int ch, int val)
{
    MusData.driver->bend(ch, val);
}

static void M_NoteOn(int ch, int note, BYTE vol, unsigned Instru, unsigned Period)
{
    M_GmBend(ch, 0x2000); /* ensure */
    MusData.driver->patch(
      ch,
      MusData.SongData.Instr[Instru]->D,
      MusData.SongData.Instr[Instru]->GM,
      MusData.SongData.Instr[Instru]->Bank & 127);
    
    MusData.PlayData.ch[ch].OrigPeriod = Period ? PERIODCONST / Period : 0;
    
    MusData.driver->noteon(
      ch,
      note,
      MusData.PlayData.ch[ch].OrigPeriod,
      vol * MusData.GlobalVolume / 255 * (MusData.SongData.Instr[Instru]->D[8] >> 2) / 63,
      vol * MusData.GlobalVolume / 255);
}

static void M_Touch(int ch, unsigned Instru, BYTE vol)
{
    MusData.driver->touch(
      ch,
      vol * MusData.GlobalVolume / 255 * (MusData.SongData.Instr[Instru]->D[8] >> 2) / 63,
      vol * MusData.GlobalVolume / 255);
}

static void M_PeriodTouch(int ch, unsigned Period)
{
    unsigned Per = Period ? PERIODCONST / Period : 0;
    double semidiff;
    int bendi;
    
    MusData.driver->hztouch(ch, Per);
    
    semidiff = 12 * log(Per / (double)MusData.PlayData.ch[ch].OrigPeriod) / log(2);
    
    /* printf("semidiff = %+.4f\n", semidiff); */
    
    bendi = 0x2000 + 0x1FFF * semidiff / 2; /* * SEMIUNIT / PITCHSENSE; */
    if(bendi < 0)bendi = 0;
    if(bendi >= 0x4000)bendi = 0x3FFF;
    
    M_GmBend(ch, bendi);
    
    /*
        a / b = 2 ^ (x / 12)   ; log
        
        log(a / b) = log(2) * (x/12)
        
        x = 12 * log(a / b) / log(2)  
    */
}

static unsigned FetchBits(unsigned ch, unsigned num)
{
    unsigned val = ReadBits(MusData.PlayData.ch[ch].patndata, MusData.PlayData.posi.bitpos[ch], num);
    MusData.PlayData.posi.bitpos[ch] += num;
    return val;
}

void MusSeekRow(unsigned Row)
{
    unsigned a;
    
    FreeUncompressedData();
    
    MusData.PlayData.posi.Pat = MusData.SongData.Orders[MusData.PlayData.posi.Ord];
    MusData.PlayData.posi.Row = Row;
    
    UncompressChannels();
    
    for(a=0; a<MusData.SongData.Hdr->ChanCount; ++a)
        MusData.PlayData.posi.bitpos[a] = 0;
    
    for(a=0; a<Row; ++a)
    {
        unsigned c;
        for(c=0; c<MusData.SongData.Hdr->ChanCount; ++c)
        {
            int k;
            k = FetchBits(c, 2);
            if(k < 3)
            {
                FetchBits(c, k*8);
                if(FetchBits(c, 1))FetchBits(c, 6);
                for(;;)
                {
                    k = FetchBits(c, 2);
                    if(!k)break;
                    FetchBits(c, 5 + (k-1)*8);
                }
            }
        }
    }    
    
    /* At this point, we are always on a new order.             */
    /* SB0 must always be assumed upon incoming to a new order. */
    MusData.PlayData.PendingSB0 = 1;
    MusData.PlayData.LoopCount  = 0;    
}

#define DISPLAYROWS       1
#define HEXDUMP           0
#define DISPLAYFRAMETICKS 0

struct HadCommands
{
    int Vibrato;
    int Tremolo;
    int Tremor;
    int Arpeggio;
    int Retrig;
    int Portamento;
    int VolSlide;
    int PerSlide;
};

struct EffectData
{
    struct MusPlayerPos ToSave;
    int Restore;
    int Cut, CutLine;
    int Break, BreakOrder;
};

static void ProcessLine(unsigned ch, struct HadCommands *Had, struct EffectData *Data)
{
    int k;
    memset(Had, 0, sizeof(*Had));
    
    MusData.PlayData.ch[ch].CurPuh = 255;

    k = FetchBits(ch, 2);
    
    switch(k)
    {
        case 3:
            /* Empty row */
            return;
        case 2:
        {
            BYTE Instru = FetchBits(ch, 8);
            printf("\r\33[%uC%02u", 12+ch*14 + 4, Instru);
            MusData.PlayData.ch[ch].CurVol = MusData.SongData.Instr[Instru]->Volume & 63;
            MusData.PlayData.ch[ch].CurInst = Instru;
            /* Bits 6 and 7 unused */
            /* fall through */
        }
        case 1:
        {
            BYTE Instru = MusData.PlayData.ch[ch].CurInst;    
            BYTE Puh    = FetchBits(ch, 8);
            MusData.PlayData.ch[ch].CurPuh = Puh;
            
            if(Puh != 255)
            {
                int r;
                BYTE Oct, Note;
                
                MusData.PlayData.ch[ch].NoteCut = 0;
                MusData.PlayData.ch[ch].NoteDelay = 1;
            
                Oct  = Puh >> 4;
                Note = Puh & 15;
                
                if(Instru==0)
                {
                    Instru=1;
                    printf("\nStrange! Instru=0.. Why?\n");
                    MusData.PlayData.ch[ch].CurInst = Instru;
                }
                
                MusData.PlayData.ch[ch].NNum = (12*Oct) + Note;
                MusData.PlayData.ch[ch].NNum += MusData.SongData.Instr[Instru]->FT;
                /* MIDI finetuning. */
    
                r = ((MusData.SongData.Instr[Instru]->D[9] >> 2) & 15);
    
                if(r)
                {
                    r = rand() % (r + 1);
                    MusData.PlayData.ch[ch].NoteDelay = r+1;
                }
                
                /* AdLib finetuning also? */
                if(MusData.SongData.Instr[Instru]->Bank & 128)
                {
                    /* Recalculate these */
                    Note = MusData.PlayData.ch[ch].NNum % 12;
                    Oct  = MusData.PlayData.ch[ch].NNum / 12;
                }
                
                MusData.PlayData.ch[ch].NNum -= 2; /* Some kind of hack. */
                
                MusData.PlayData.ch[ch].NextPeriod =
                    MakePeriod(Note, Oct, MusData.SongData.Instr[Instru]->C4Spd);

#if DISPLAYROWS
                printf("\r\33[%uC%c%c%u",
                     12+ch*14,
                     "CCDDEFFGGAAB"[Note],
                     "-#-#--#-#-#-"[Note],
                     Oct);
#endif
            }
        }
        case 0:
            /* no note, no instrument */
    }
        
    if(FetchBits(ch, 1))
    {
        BYTE Vol = FetchBits(ch, 6);
#if DISPLAYROWS
     /* C-4 10 63 A06|
        12345678901234
        01234567890
        */
    
        printf("\r\33[%uC%02u", 12+ch*14 + 7, Vol);
#endif
            
        MusData.PlayData.ch[ch].CurVol = Vol;
        MusData.PlayData.ch[ch].NoteDelay = 1;
    }
    
    while((k = FetchBits(ch, 2)) > 0)
    {
        BYTE Cmd;
        WORD Info = 0;
        
        switch(k)
        {
            case 3:
                Info = FetchBits(ch, 16);
                Cmd = FetchBits(ch, 5);
                break;
            case 2:
                Info = FetchBits(ch, 8);
                /* fallthrough */
            case 1:
                Cmd = FetchBits(ch, 5);
                break;
        }
        
        printf("\r\33[%uC%c%02X", 12+ch*14 + 10, Cmd|64, Info);

        switch(Cmd | 64)
        {
            case 'A':
                MusData.PlayData.FrameCount = Info;
                break;
            case 'B':
                Data->Break = 1;
                /* The order is in hexadecimal format */
                Data->BreakOrder = Info;
                break;
            case 'C':
                Data->Cut = 1;
                /* The line is in bcd format */
                Data->CutLine =
                     ((Info)    & 15)
                   + ((Info>>4) & 15) * 10
                   + ((Info>>8) & 15) * 100
                   + ((Info>>12)& 15) * 1000;
                break;
            case 'D':
            case 'K':
                if(!Info)Info = MusData.PlayData.ch[ch].DxxDefault;
                else MusData.PlayData.ch[ch].DxxDefault = Info;

                if(Info==0xFF)MusData.PlayData.ch[ch].CurVol = 63;
                                                         /* DFF, maximize */
                else if(Info > 0xF0)                     /* DFx           */
                {
                    MusData.PlayData.ch[ch].VolSlide = -(Info&0x0F);
                    MusData.PlayData.ch[ch].FineVolSlide = 1;
                }
                else if((Info&0x0F)==0x0F && (Info&0xF0)) /* DxF */
                {
                    MusData.PlayData.ch[ch].VolSlide = Info>>4;
                    MusData.PlayData.ch[ch].FineVolSlide = 1;
                }
                else if(!Info);                           /* D00 */
                else if(!(Info&0xF0))                     /* D0x */
                {
                    MusData.PlayData.ch[ch].VolSlide = -(Info&0x0F);
                    MusData.PlayData.ch[ch].FineVolSlide = 0;
                }
                else if(!(Info&0x0F))                     /* Dx0 */
                {
                    MusData.PlayData.ch[ch].VolSlide = Info>>4;
                    MusData.PlayData.ch[ch].FineVolSlide = 0;
                }
                Had->VolSlide = 1;
                if(Cmd=='K') { Info=0; goto CmdH; }
                if(Cmd=='L') { Info=0; goto CmdG; }
                break;
            
            case 'E':
            case 'F':
                MusData.PlayData.ch[ch].FinePerSlide = 0;
                if(!Info)Info = MusData.PlayData.ch[ch].ExxDefault;
                else MusData.PlayData.ch[ch].ExxDefault = Info;
                if(Info >= 0xF0)MusData.PlayData.ch[ch].FinePerSlide=1, Info&=15;
                else if(Info >= 0xE0)MusData.PlayData.ch[ch].FinePerSlide=2, Info&=15;
                MusData.PlayData.ch[ch].PerSlide = (Cmd=='F') ? -Info : Info;
                Had->PerSlide = 1;
                break;
            
            case 'G':
CmdG:           if(!Info)Info = MusData.PlayData.ch[ch].GxxDefault;
                else MusData.PlayData.ch[ch].GxxDefault = Info;
                MusData.PlayData.ch[ch].Portamento = Info;
                Had->Portamento = 1;
                break;
            
            case 'H':
            case 'U':
CmdH:           if(!(Info & 0x0F) || !(Info >> 4))Info=MusData.PlayData.ch[ch].HxxDefault;
                else MusData.PlayData.ch[ch].HxxDefault=Info;
                
                MusData.PlayData.ch[ch].Vibrato = Info;
                MusData.PlayData.ch[ch].FineVibra = (Cmd == 'U');
                Had->Vibrato = 1;
                break;
            
            case 'I':
                if(!Info)Info=MusData.PlayData.ch[ch].IxxDefault;
                else MusData.PlayData.ch[ch].IxxDefault=Info;
                MusData.PlayData.ch[ch].Tremor = Info;
                Had->Tremor = 1;
                break;

            case 'J':
                if(!Info)Info=MusData.PlayData.ch[ch].JxxDefault;
                else MusData.PlayData.ch[ch].JxxDefault=Info;
                MusData.PlayData.ch[ch].Arpeggio = Info;
                Had->Arpeggio = 1;
                break;

            case 'M':
                if(!Info)Info = MusData.PlayData.ch[ch].MxxDefault;
                MusData.PlayData.ch[ch].MxxDefault = Info;
                
                /* MBx is reverb.
                 * MCx is chorus.
                 * MAx is supported also, because I've accidentally
                 *        wrote some songs using it... //Bisqwit
                 */
                
                if((Info>>4)==0xB
                || (Info>>4)==0xA)
                {
                    /* Reverb is sticky. */
                    M_MidiCtrl(ch, 91, (Info&15)<<3);
                }
                else if((Info>>4)==0xC)
                {
                    /* Chorus is sticky. */
                    M_MidiCtrl(ch, 93, (Info&15)<<3);
                }
                break;

            case 'N':
                /* Bend is not sticky.
                 * It will automatically reset upon next note.
                 * FIXME: This should be done AFTER NoteDelay!
                 */
                
                /* Bend limits: 0x0000 - 0x3FFF
                 * 0x80 << 6 = 0x2000
                 * 0x00 << 6 = 0x0000
                 * 0xFF << 6 = 0x3FE0 or similar
                 */
                M_GmBend(ch, (unsigned)Info << 6);
                break;
                
            case 'P':
                if(Info)MusData.PlayData.ArpegSpd = Info;
                break;

            case 'Q':
                if(Info)MusData.PlayData.ch[ch].Retrig = Info;
                Had->Retrig = 1;
                break;

            case 'R':
                if(!(Info & 0x0F) || !(Info >> 4))Info=MusData.PlayData.ch[ch].RxxDefault;
                else MusData.PlayData.ch[ch].RxxDefault=Info;
                MusData.PlayData.ch[ch].Tremolo = Info;
                Had->Tremolo = 1;
                break;
            
            case 'S':
            {
                BYTE a, b;
             
                if(!Info)Info = MusData.PlayData.ch[ch].SxxDefault; /* Notice */
                MusData.PlayData.ch[ch].SxxDefault = Info;          /* Notice */
                
                a = Info >> 4;
                b = Info & 15;
                
                switch(a)
                {
                    case 0x3:
                        MusData.PlayData.ch[ch].VibraForm = b;
                        break;

                    case 0x4:
                        MusData.PlayData.ch[ch].TremoForm = b;
                        break;
                        
                    case 0xB:
                        if(!b)
                        {
                            /* ToSave is what we had on the beginning of line */
                            MusData.PlayData.saved = Data->ToSave;
                        }
                        else
                        {
                            /* If there was no active loop going on */
                            if(MusData.PlayData.LoopCount == 0)
                            {
                                /* Then set it */
                                MusData.PlayData.LoopCount = b;
                            }
                            else
                            {
                                /* Else decrease the ongoing counter */
                                --MusData.PlayData.LoopCount;
                            }
                            
                            /* If there is still something to loop */
                            if(MusData.PlayData.LoopCount)
                            {
                                /* Then we'll go back when we can */
                                Data->Restore = 1;
                            }
                            else
                            {
                                /* Else we're done, and we must do SB0 at next line. */
                                MusData.PlayData.PendingSB0 = 1;
                            }
                        }
                        break;
                    case 0xC:
                        /* Mitkhn ideaa olisi komennossa SC0... */
                        MusData.PlayData.ch[ch].NoteCut = b + 1;
                        break;
                    case 0xD:
                        /* Mitkhn ideaa olisi komennossa SD0... */
                        MusData.PlayData.ch[ch].NoteDelay = b + 1;
                        break;
                    case 0xE:
                        /* Mitkhn ideaa olisi komennossa SE0... */
                        MusData.PlayData.PatternDelay = b;
                        break;
                }
                break;
            }
            
            case 'T':
                MusData.PlayData.Tempo = Info;
                break;
            
            case 'X':
                /* Pan is sticky. */
                M_Pan(ch, Info);
                break;
        }
    }
}

static void PlayLine(void)
{
    struct EffectData Data;
    int ch;

    if(MusData.PlayData.PatternDelay)
    {
        unsigned a;
        /* Set empty row, because we don't want *
         * to retrig the notes while doing SEx. */
        for(a=0; a<MAXCHN; MusData.PlayData.ch[a++].CurPuh = 255);
        --MusData.PlayData.PatternDelay;
        return;
    }
    
    memset(&Data, 0, sizeof(Data));
    
    Data.ToSave = MusData.PlayData.posi;

    if(MusData.PlayData.PendingSB0)
    {
        /* Do saving at the pattern first line! */
        MusData.PlayData.saved = Data.ToSave;
        MusData.PlayData.PendingSB0 = 0;
    }
    
#if DISPLAYROWS
    printf("o%02X p%02u r%02u|",
        MusData.PlayData.posi.Ord,
        MusData.PlayData.posi.Pat,
        MusData.PlayData.posi.Row);
    
    for(ch=0; ch<MusData.SongData.Hdr->ChanCount; ++ch)
    {
        printf("\33[1;30m... .. .. .00\33[34m|\33[37m");
           /*..............|01234567890123 */
    }
#endif
    
    for(ch=0; ch<MusData.SongData.Hdr->ChanCount; ++ch)
    {
        struct HadCommands Had;
        
        ProcessLine(ch, &Had, &Data);
        
        if(!Had.Vibrato)MusData.PlayData.ch[ch].Vibrato = 0;
        if(!Had.Tremolo)MusData.PlayData.ch[ch].Tremolo = 0;
        if(!Had.Tremor)
        {
            MusData.PlayData.ch[ch].TremoClk = 0;
            MusData.PlayData.ch[ch].Tremor   = 0;
        }
        if(!Had.Arpeggio)MusData.PlayData.ch[ch].Arpeggio = 0;
        if(!Had.Retrig)MusData.PlayData.ch[ch].Retrig  = 0;
        if(!Had.Portamento)MusData.PlayData.ch[ch].Portamento = 0;
        if(!Had.VolSlide)MusData.PlayData.ch[ch].VolSlide = 0;
        if(!Had.PerSlide)MusData.PlayData.ch[ch].PerSlide = 0;
        
        if(!Had.PerSlide && !Had.Portamento && !Had.Vibrato && !Had.Tremor)
            M_PeriodTouch(ch, MusData.PlayData.ch[ch].ST3Period);
    }

#if DISPLAYROWS
    printf("\33[m\n");
#endif

    if(Data.Restore)
    {
        /* If got SBx _now_ */
        MusData.PlayData.posi = MusData.PlayData.saved;
        return;
    }
    
    if(!MusData.PlayData.PatternDelay)
        ++MusData.PlayData.posi.Row;

    if(Data.Break)
    {
        MusData.PlayData.posi.Ord = Data.BreakOrder;
        
        /* Bxx, hyptn mrttyyn orderiin.
         * Jos oli mys Cxx, hyptn mrtylle riville.
         */
        
        printf("B%02X %d? C%02u\n", Data.BreakOrder, Data.Cut, Data.CutLine);
        MusSeekRow(Data.Cut ? Data.CutLine : 0);
    }
    else if(Data.Cut)
    {
        /* Cxx, hyptn seuraavaan orderiin mrtylle riville. */
        /* ++ -orderit yms pitisi loistaa poissaolollaan, joten tm toimii */
        MusData.PlayData.posi.Ord = (MusData.PlayData.posi.Ord+1) % (MusData.SongData.Hdr->OrdNum);
        
        printf("C%02u\n", Data.CutLine);
        MusSeekRow(Data.CutLine);
    }
    else if(MusData.PlayData.posi.Row >= MusData.SongData.Patn[MusData.PlayData.posi.Pat]->RowCount)
    {
        /* Jos patterni loppui kesken, mennn seuraavaan orderiin, riville 0. */
        
        MusData.PlayData.posi.Ord = (MusData.PlayData.posi.Ord+1) % (MusData.SongData.Hdr->OrdNum);
        
        MusSeekRow(0);
    }
}

static short vibval(volatile WORD *const index, BYTE mode)
{
    static const short sintab[32] =
    {
        0x00,0x18,0x31,0x4a,0x61,0x78,0x8d,0xa1,
        0xb4,0xc5,0xd4,0xe0,0xeb,0xf4,0xfa,0xfd,
        0xff,0xfd,0xfa,0xf4,0xeb,0xe0,0xd4,0xc5,
        0xb4,0xa1,0x8d,0x78,0x61,0x4a,0x31,0x18
    };
    static const short ramptab[64] =
    {
        0,-248,-240,-232,-224,-216,-208,-200,-192,
        -184,-176,-168,-160,-152,-144,-136,-128,-120,
        -112,-104,-96,-88,-80,-72,-64,-56,-48,-40,
        -32,-24,-16,-8,0,8,16,24,32,40,48,56,64,72,
        80,88,96,104,112,120,128,136,144,152,160,
        168,176,184,192,200,208,216,224,232,240,248
    };

    if(mode & 4)
        *index &= 63;
    else if(*index & 64)
        *index = 0;
    
    switch(mode & 3)
    {
        case 0:                              /* sin    */
        {
            short tmp = sintab[*index & 31];
            if(*index & 32)tmp = -tmp;
            return tmp;
        }
        case 1: return ramptab[*index];      /* ramp   */
        case 2: return (*index&32) ? 0 : 255;/* square */
        case 3:                              /* random */
        {
            static DWORD key;
            key = (key * 0x8088401 + 1);
            return sintab[*index] + (key >> 24);
        }
    }
    return 0; /* unreachable code */
}

static void Vibrato(unsigned ch)
{
    /**** Hxy and Kxy and Uxy ****/
    
    if(MusData.PlayData.ch[ch].ST3Period && MusData.PlayData.ch[ch].Vibrato)
    {
        BYTE VibSpd   = MusData.PlayData.ch[ch].Vibrato >> 4;
        BYTE VibDepth = MusData.PlayData.ch[ch].Vibrato & 15;

        int Val = vibval(&MusData.PlayData.ch[ch].VibPos, MusData.PlayData.ch[ch].VibraForm) * VibDepth;
        Val >>= MusData.PlayData.ch[ch].FineVibra ? 7 : 5; /* old-style: 6 and 4 */
        
        M_PeriodTouch(ch, MusData.PlayData.ch[ch].ST3Period + Val);
        
        MusData.PlayData.ch[ch].VibPos += VibSpd;
    }
}

static void Tremolo(unsigned ch)
{
    /**** Rxy ****/
    
    if(MusData.PlayData.ch[ch].CurVol && MusData.PlayData.ch[ch].Tremolo)
    {
        BYTE TremoSpd   = MusData.PlayData.ch[ch].Tremolo >> 4;
        BYTE TremoDepth = MusData.PlayData.ch[ch].Tremolo & 15;

        /* Rxy: 
                 Tremolo amplitude: y*(ticks-1) semitones
                 Tremolo speed:     x*ticks/64 cycles in second
                 Ticks is the value set by Axx -command.
        */

        int Val = vibval(&MusData.PlayData.ch[ch].TremoPos, MusData.PlayData.ch[ch].TremoForm) * TremoDepth;
        int Vol = MusData.PlayData.ch[ch].CurVol + (Val >> 7);
        
        if(Vol < 0)Vol = 0;
        if(Vol > 63)Vol = 63;

        M_Touch(ch, MusData.PlayData.ch[ch].CurInst, Vol);
        
        MusData.PlayData.ch[ch].TremoPos += TremoSpd;
    }
}

static void Retrig(unsigned ch)
{
    /**** Qxy ****/
    
    if(MusData.PlayData.ch[ch].Retrig)
    {
        unsigned space = MusData.PlayData.ch[ch].Retrig & 15;
        unsigned fade  = MusData.PlayData.ch[ch].Retrig >> 4;
        
        if(!(MusData.PlayData.Frame % space))
        {
            if(MusData.PlayData.ch[ch].CurVol > fade)
             MusData.PlayData.ch[ch].CurVol -= fade;
         else
             MusData.PlayData.ch[ch].CurVol = 0;
            
            M_NoteOff(ch);
            M_NoteOn(ch, MusData.PlayData.ch[ch].NNum,
                         MusData.PlayData.ch[ch].CurVol,
                         MusData.PlayData.ch[ch].CurInst,
                         MusData.PlayData.ch[ch].ST3Period);
        }
    }
}

static void VolSlide(unsigned ch)
{
    /**** Dxy and Kxy and Lxy ****/
    
    if(MusData.PlayData.ch[ch].VolSlide)
    {
        int Vol = MusData.PlayData.ch[ch].CurVol;
        
        if(!MusData.PlayData.Frame || !MusData.PlayData.ch[ch].FineVolSlide)
            Vol += MusData.PlayData.ch[ch].VolSlide;
            
        if(Vol < 0)Vol = 0;
        if(Vol > 63)Vol = 63;

        MusData.PlayData.ch[ch].CurVol = Vol;
        
        M_Touch(ch,
          MusData.PlayData.ch[ch].CurInst,
          MusData.PlayData.ch[ch].CurVol);

        if(MusData.PlayData.ch[ch].FineVolSlide && !MusData.PlayData.PatternDelay)
            MusData.PlayData.ch[ch].VolSlide = 0;
    }
}

static void Tremor(unsigned ch)
{
    /**** Ixy ****/
    
    if(MusData.PlayData.ch[ch].Tremor)
    {
        BYTE on = MusData.PlayData.ch[ch].Tremor >> 4; /* These are 0-based! 0 on means */
        BYTE off= MusData.PlayData.ch[ch].Tremor & 15; /* that note is on for one tick  */
        
        /* Because we didn't call this function at the first tick,
         * TremoClk has already 1 although it seems like 0.
         */
        /* This _is_ correct. */
        
        if(MusData.PlayData.ch[ch].TremoClk >= on)
        {
            /* Been on enough, now we switch it off */
            M_Touch(ch,
              MusData.PlayData.ch[ch].CurInst,
              0);
        }
        
        if(++MusData.PlayData.ch[ch].TremoClk > on+off)
        {
            /* Was off enough, we switch it back on */
            MusData.PlayData.ch[ch].TremoClk = 0;
            M_Touch(ch,
              MusData.PlayData.ch[ch].CurInst,
              MusData.PlayData.ch[ch].CurVol);
        }
    }        
}

static void PerSlide(unsigned ch)
{
    /**** Exy and Fxy ****/

    if(MusData.PlayData.ch[ch].ST3Period && MusData.PlayData.ch[ch].PerSlide)
    {
        if(!MusData.PlayData.Frame || !MusData.PlayData.ch[ch].FinePerSlide)
        {
            int tmp = MusData.PlayData.ch[ch].PerSlide
                    * ((MusData.PlayData.ch[ch].FinePerSlide == 2) ? 1 : 4);
            MusData.PlayData.ch[ch].ST3Period += tmp;
              
            M_PeriodTouch(ch, MusData.PlayData.ch[ch].ST3Period);
        }

        if(MusData.PlayData.ch[ch].FinePerSlide && !MusData.PlayData.PatternDelay)
            MusData.PlayData.ch[ch].PerSlide = 0;
    }
}

static void Portamento(unsigned ch)
{
    /**** Gxx and Lxy ****/

    if(MusData.PlayData.ch[ch].ST3Period && MusData.PlayData.ch[ch].Portamento)
    {
        int change = MusData.PlayData.ch[ch].Portamento * 4, done=0;
        if(MusData.PlayData.ch[ch].ST3Period < MusData.PlayData.ch[ch].NextPeriod)
        {
            MusData.PlayData.ch[ch].ST3Period += change;
            if(MusData.PlayData.ch[ch].ST3Period >= MusData.PlayData.ch[ch].NextPeriod)done=1;
        }
        else
        {
            MusData.PlayData.ch[ch].ST3Period -= change;
            if(MusData.PlayData.ch[ch].ST3Period <= MusData.PlayData.ch[ch].NextPeriod)done=1;
        }
        if(done)
        {
            MusData.PlayData.ch[ch].ST3Period = MusData.PlayData.ch[ch].NextPeriod;
            MusData.PlayData.ch[ch].Portamento = 0;
        }
        
        M_PeriodTouch(ch, MusData.PlayData.ch[ch].ST3Period);
    }
}

static void Arpeggio(unsigned ch)
{
    /* Arpeggio has its own timer. */
    DWORD ArpegCounter = (unsigned long)(MusData.PlayData.ArpegSpd & 127) * CLKMUL / Rate;
    
    /**** Jxy ****/
    if(MusData.PlayData.ch[ch].ST3Period && MusData.PlayData.ch[ch].Arpeggio)
    {
        for(MusData.PlayData.ch[ch].ArpegClk += ArpegCounter;
            MusData.PlayData.ch[ch].ArpegClk >= CLKMUL; /* was overflow? */
            MusData.PlayData.ch[ch].ArpegClk -= CLKMUL)
        {
            int notemp, Note, NoteAdd;
            BYTE Instru = MusData.PlayData.ch[ch].CurInst;
            
            Note = MusData.PlayData.ch[ch].NNum; // is oct*12 + note
            
            switch((MusData.PlayData.ch[ch].ArpegCnt++) % (3-(MusData.PlayData.ArpegSpd>>7)))
            {
                case 0:
                    // nothing
                    break;
                case 1:
                    Note += MusData.PlayData.ch[ch].Arpeggio & 15;
                    break;
                default:
                    Note += MusData.PlayData.ch[ch].Arpeggio >> 4;
            }
            
            notemp = Note;
            
            // FIXME: Does it really work on MPU?
            notemp += 2; // some kind of hack

#if 1
            if(!(MusData.SongData.Instr[Instru]->Bank & 128))
            {
                notemp -= MusData.SongData.Instr[Instru]->FT - 2;
                /* Remove finetuning effect for AdLib */
            }
#endif
            
        //  M_NoteOff(ch);
            M_NoteOn(ch, Note, MusData.PlayData.ch[ch].CurVol, Instru,
                     MakePeriod(notemp%12, notemp/12,
                                MusData.SongData.Instr[Instru]->C4Spd));
        }
    }
}

static void FrameTick(void)
{
    unsigned ch;

#if DISPLAYFRAMETICKS
    printf("\33[32mFrame Tick: Ord %02X Pat %02u Row %02u\33[m\n",
        MusData.PlayData.posi.Ord,
        MusData.PlayData.posi.Pat,
        MusData.PlayData.posi.Row);
#endif

    for(ch=0; ch<MAXCHN; ch++)
    {
        if(MusData.PlayData.ch[ch].NoteCut > 0)
        {
            if(!--MusData.PlayData.ch[ch].NoteCut)
                M_NoteOff(ch);
        }

        if(MusData.PlayData.ch[ch].NoteDelay > 0)
        {
            /* Vibrato - don't use */
            /* M_MidiCtrl(CurChn, 1, 0); */
            
            if(!--MusData.PlayData.ch[ch].NoteDelay)
                switch(MusData.PlayData.ch[ch].CurPuh)
                {
                    case 255:   /* Tyhj rivi */
                        if(MusData.PlayData.ch[ch].CurInst)
                            M_Touch(ch, MusData.PlayData.ch[ch].CurInst,
                                        MusData.PlayData.ch[ch].CurVol);
                        break;
                    case 254:   /* ^^ -rivi */
                        M_NoteOff(ch);
                        break;
                    default:    /* Rivi jolla nuotti */
                        M_NoteOff(ch);
                        
                        if(!MusData.PlayData.ch[ch].Portamento)
                        {
                            MusData.PlayData.ch[ch].ST3Period = MusData.PlayData.ch[ch].NextPeriod;
                        }
                        
                        M_NoteOn(ch, MusData.PlayData.ch[ch].NNum,
                                     MusData.PlayData.ch[ch].CurVol,
                                     MusData.PlayData.ch[ch].CurInst,
                                     MusData.PlayData.ch[ch].ST3Period);
                }
                /* FIXME:
                   When PatternDelay is used, NoteDelay
                   should retrigger the same note again
                   and again.
                    !MusData.PlayData.PatternDelay
                */
        }
        
        /* These are done once per frame */
        VolSlide(ch);
        PerSlide(ch);

        /* Some operations must not be done at the first frame */
        if(MusData.PlayData.Frame)
        {
            Tremor(ch);
            Retrig(ch);
            Portamento(ch);
            Vibrato(ch);
            Tremolo(ch);
        }
    }
}

static void AccurateEffects(void)
{
    unsigned ch;
    for(ch=0; ch<MAXCHN; ++ch)
    {
        Arpeggio(ch);
    }
}

/* Handles one timer tick */
/* Called from m_irq.c    */
void TimerHandler(void)
{
    if(MusData.Paused || !MusData.driver)return;
    
    for(MusData.PlayData.FrameClk += MusData.PlayData.FrameSpd;
        MusData.PlayData.FrameClk >= CLKMUL;
        MusData.PlayData.FrameClk -= CLKMUL)
    {
        /* tick! */
        
#if 0
        printf("Clock = %08X, Speed = %08X (T%02X A%02X)\n",
          MusData.PlayData.FrameClk,
          MusData.PlayData.FrameSpd,
          MusData.PlayData.Tempo,
          MusData.PlayData.FrameCount);
#endif

        if(++MusData.PlayData.Frame >= MusData.PlayData.FrameCount)
        {
            ++MusData.NextRow;            
            PlayLine();
            
            MusData.PlayData.Frame  = 0;
            
            /* By definition, rows played in minute is 24 * tempo / ticks.
            
               Therefore frames per minute is 24 * tempo
               
               Frames per second = 24 * tempo / 60
                                 = 2 * tempo / 5
            */
            
            MusData.PlayData.FrameSpd = (unsigned long)MusData.PlayData.Tempo * 2 * CLKMUL / 5 / Rate;
            
#if 0
            printf("Rows per second is %.2f\n",
                2 * MusData.PlayData.Tempo / 5.0 / MusData.PlayData.FrameCount);
#endif
        }
        
        /* Frame has ticked, let's do something */
        
        ++MusData.NextFrame;
        FrameTick();
    }
        
    ++MusData.NextTick;
    
    AccurateEffects();
}

static void DumpPeriods(void)
{
    int oct, note;
    
    for(note=0; note<12; ++note)
        printf("\t%d", note);
    printf("\n");
    for(oct=0; oct<8; ++oct)
    {
        printf("%d\t", oct);
        for(note=0; note<12; ++note)
            printf("%d\t", MakePeriod(note, oct, 8363));
        printf("\n");
    }
}

void StartS3M(const BYTE *s)
{
    unsigned a, b;

    struct MusDriver *tmp = MusData.driver;

    StopS3M();
    
    memset((void *)&MusData, 0, sizeof(MusData));
    MusData.driver = tmp;
    
    /*debug- DumpPeriods();*/
    
    /* Pointers are nice :) */
    MusData.SongData.Hdr = (const struct InternalHdr *)__FILE__" v"VERSION"(C) 1992,2000 Bisqwit (bisqwit@iki.fi)";
    /* Please don't modify the text above. It would be nice to   *
     * see my name inside games made by people unknown to me :-) */

    MusData.SongData.Playing = s;
    if(s[1]>2)for(;*s++;); /* Skip possible song name */

    MusData.SongData.Hdr = (const struct InternalHdr *)s;
    s += sizeof(struct InternalHdr);
    
    MusData.SongData.HaveINames    = (MusData.SongData.Hdr->PatNum  & 512) == 512;
    MusData.SongData.LinearMidiVol = (MusData.SongData.Hdr->InSpeed & 128) == 128;
    
    MusData.PlayData.Tempo      = MusData.SongData.Hdr->InTempo;
    MusData.PlayData.FrameCount = MusData.SongData.Hdr->InSpeed & 127;
    MusData.PlayData.ArpegSpd   = 50;
    MusData.GlobalVolume = 255;
    
    MusData.SongData.Orders = (const BYTE *)s;
    s += MusData.SongData.Hdr->OrdNum;
    
    for(a=0; a<MusData.SongData.Hdr->InsNum; a++)
    {
        const struct InternalSample *tmp = (const struct InternalSample *)s;
        s += sizeof(*tmp);
        
        if(MusData.SongData.HaveINames)
        {
            MusData.SongData.InsName[a] = (const char *)s;
            s += 1 + strlen(s);
        }
        else
            MusData.SongData.InsName[a] = "";
       
        MusData.SongData.Instr[a+1] = tmp;
    }
    
    /* This disables some bugs. */
    /* MusData.SongData.Instr[0] = MusData.SongData.Instr[1]; */
    MusData.SongData.Instr[0] = NULL;
    /* We want to FIND those bugs! */

    b = MusData.SongData.Hdr->PatNum & 511;
    for(a = 0; a < b; a++)
    {
        unsigned c;
        const struct InternalPattern *p = (const struct InternalPattern *)s;
        
        MusData.SongData.Patn[a] = p;
        
        s += sizeof(*p);
        
        for(c=0; c<MusData.SongData.Hdr->ChanCount; ++c)
        {
            const struct InternalChanData *q = (const struct InternalChanData *)s;
            if(!c)
                MusData.SongData.Chans[a] = q;
        
            s += sizeof(*q) + ((q->BitCount+7) >> 3);
        }
    }
    
    for(a=0; a<MAXCHN; ++a)
    {
        /* Set pitchbend sensitivity to max! */
        M_MidiCtrl(a, 6, PITCHSENSE);
    }
    
    SetPause(1);

    MusSeekOrd(0);
    
    SetPause(0);
}

/* From m_arch.c */
extern int FromArchive;
extern char *Archived;

void StopS3M(void)
{
    FreeUncompressedData();
    
    SetPause(1);
    ReleaseINT();
    
    if(Archived && !FromArchive)
    {
        free(Archived);
        Archived = NULL;
    }
}

void SetPause(int state)
{
    if(state == MusData.Paused)return;
    MusData.Paused = state;
    
    if(MusData.driver)
        MusData.driver->reset();

    if(state) /* pause */
    {
        /* actually nothing to do */
    }
    else      /* unpause */
    {
        InstallINT();
    }
}

void PauseS3M(void)
{
    if(MusData.Paused)
        SetPause(0);
    else
        SetPause(1);
}

void MusSeekOrd(unsigned Num)
{    
    int p = MusData.Paused;
    
    SetPause(1);
    
    /* The player must not modify these at same time */
    MusData.PlayData.PatternDelay = 0;
    MusData.PlayData.FrameSpd   = CLKMUL-1; /* Hurry first. */
    MusData.PlayData.Frame      = MusData.PlayData.FrameCount-1;
    MusData.PlayData.posi.Ord   = Num;
    
    MusSeekRow(0);
    
    SetPause(p);
}

void MusSeekNextOrd(void)
{
    unsigned limit = MusData.SongData.Hdr->OrdNum - 1;
    if(MusData.PlayData.posi.Ord < limit)
        MusSeekOrd(MusData.PlayData.posi.Ord + 1);
}

void MusSeekPrevOrd(void)
{
    if(MusData.PlayData.posi.Ord > 0)MusSeekOrd(MusData.PlayData.posi.Ord - 1);
}
